from . import utils
import itertools
from os import path
import pandas as pd
import ast
import math
import logging
from .cutils import get_initial_isotopes, checking_cos_correlation_for_carbon, split_peaks, split_peaks_old, detect_hills, process_hills, get_and_calc_values_for_cos_corr, cos_correlation, get_and_calc_apex_intensity_and_scan

logger = logging.getLogger(__name__)


def process_file(args):

    input_mzml_path = args['file']
    basename_mzml = path.basename(input_mzml_path)

    if args['o']:
        output_file = args['o']
    else:
        output_file = path.splitext(input_mzml_path)[0]\
            + path.extsep + 'features.tsv'

    df1_features = pd.read_table(output_file)
    # df1_features['mono_hills_scan_lists'] = df1_features['mono_hills_scan_lists'].apply(lambda x: ast.literal_eval(','.join(x.split(' '))))
    # df1_features['mono_hills_scan_lists'] = df1_features['mono_hills_scan_lists'].apply(lambda x: x.replace('[', '').replace(']', '').split(' '))#ast.literal_eval(','.join(x.split(' '))))
    df1_features['mono_hills_scan_lists'] = df1_features['mono_hills_scan_lists'].apply(lambda x: ast.literal_eval(x))
    df1_features['mono_hills_scan_sets'] = df1_features['mono_hills_scan_lists'].apply(lambda x: set(x))
    df1_features['mono_hills_intensity_list'] = df1_features['mono_hills_intensity_list'].apply(lambda x: ast.literal_eval(x))

    def calc_idict(x):
        hill_idict_1 = dict()
        for scan_id_val, intensity_val in zip(x['mono_hills_scan_lists'], x['mono_hills_intensity_list']):
            hill_idict_1[scan_id_val] = intensity_val
        return hill_idict_1

    def calc_sqrt_of_i(x):
        hill_sqrt_of_i_1 = math.sqrt(sum(v**2 for v in x['hill_idict_1'].values()))
        return hill_sqrt_of_i_1

    df1_features['hill_idict_1'] = df1_features.apply(calc_idict, axis=1)
    df1_features['hill_sqrt_of_i_1'] = df1_features.apply(calc_sqrt_of_i, axis=1)

    logger.debug(df1_features.head())


    if args['mgf']:
        outmgf_name = args['mgf']
    else:
        outmgf_name = path.splitext(input_mzml_path)[0]\
            + path.extsep + 'mgf'

    outmgf = open(outmgf_name, 'w')
    t_i = 1


    data_for_analyse = utils.process_mzml_dia(args)

    isolation_target_func = lambda x: x['precursorList']['precursor'][0]['isolationWindow']['isolation window target m/z']
    isolation_window_func = lambda x: x['precursorList']['precursor'][0]['isolationWindow']['isolation window lower offset']

    #Process faims
    faims_set = set([z.get('FAIMS compensation voltage', None) for z in data_for_analyse])
    if any(z is not None for z in faims_set):
        logger.info('Detected FAIMS values: %s', faims_set)

    #Process windows
    windows_set = set([isolation_target_func(z) for z in data_for_analyse])
    logger.info('Detected windows values: %s', windows_set)

    RT_dict = dict()

    for (faims_val, window_val) in itertools.product(faims_set, windows_set):

        isolation_window = False
        data_cur_id = 0

        data_for_analyse_tmp = []
        for z in data_for_analyse:
            if faims_val is None or z['FAIMS compensation voltage'] == faims_val:
                if isolation_target_func(z) == window_val:
                    data_for_analyse_tmp.append(z)
                    RT_dict[data_cur_id] = float(z['scanList']['scan'][0]['scan start time'])
                    data_cur_id += 1

                    if isolation_window is False:
                        isolation_window = isolation_window_func(z)

        hill_mass_accuracy = args['diahtol']
        max_mz_value = 0
        for z in data_for_analyse_tmp:
            max_mz_value = max(max_mz_value, z['m/z array'].max())

        mz_step = hill_mass_accuracy * 1e-6 * max_mz_value

        #Process ion mobility

        if all('ignore_ion_mobility' not in z for z in data_for_analyse_tmp):
            logger.debug('%d %d', sum(('ignore_ion_mobility' in z for z in data_for_analyse_tmp)), len(data_for_analyse_tmp))
            utils.centroid_pasef_data(data_for_analyse_tmp, args, mz_step)
        else:
            args['paseftol'] = False

        paseftol = args['paseftol']

        hills_dict = detect_hills(data_for_analyse_tmp, args, mz_step, paseftol, dia=True)


        mz_step = isolation_window

        hills_dict = process_hills(hills_dict, data_for_analyse_tmp, mz_step, paseftol, args, dia=True)

        num_hills = len(set(hills_dict['hills_idx_array']))

        logger.info('All data converted to %d hills...', num_hills)


        if num_hills:

            mz_tol = isolation_window

            cnt_tmp = 0
            for ms1_mz, ms1_I, ms1_RT, ms1_ch, hill_scans_1, hill_scans_1_list, ms1_intensities, hill_idict_1, hill_sqrt_of_i_1 in df1_features[['mz', 'intensityApex', 'rtApex','charge', 'mono_hills_scan_sets', 'mono_hills_scan_lists', 'mono_hills_intensity_list', 'hill_idict_1', 'hill_sqrt_of_i_1']].values:

                hill_scans_1_list_first, hill_scans_1_list_last = hill_scans_1_list[0], hill_scans_1_list[-1]

                cnt_tmp += 1
                tmp_candidates = []

                if window_val - isolation_window <= ms1_mz <= window_val + isolation_window:

                    for idx_2, hill_idx_2 in enumerate(hills_dict['hills_idx_array_unique']):
                    # for idx_2 in hills_dict['hills_mz_median_fast_dict'][m_to_check_fast]:

                        # if paseftol is False or idx_2 in hills_dict['hills_im_median_fast_dict'][im_to_check_fast]:

                            # series_2 = idx_2

                        hill_scans_2_list = hills_dict['hills_scan_lists'][idx_2]
                        hill_scans_2_list_first, hill_scans_2_list_last = hill_scans_2_list[0], hill_scans_2_list[-1]

                        if not hill_scans_1_list_last < hill_scans_2_list_first and not hill_scans_2_list_last < hill_scans_1_list_first:
                            hill_scans_2 = hills_dict['hills_scan_sets'][idx_2]
                            if len(hill_scans_1.intersection(hill_scans_2)) >= 1:


                                hills_dict, hill_idict_2, hill_sqrt_of_i_2 = get_and_calc_values_for_cos_corr(hills_dict, idx_2)


                                # cos_cor_RT = 1.0
                                cos_cor_RT = cos_correlation(hill_scans_1, hill_idict_1, hill_sqrt_of_i_1, hill_scans_2, hill_idict_2, hill_sqrt_of_i_2)


                                if cos_cor_RT >= 0.6:

                                    hill_mz_2 = hills_dict['hills_mz_median'][idx_2]

                                    # hill_idx_2 = hills_dict['hills_idx_array_unique'][idx_2]

                                    hills_dict, _, _ = get_and_calc_apex_intensity_and_scan(hills_dict, idx_2)

                                    local_isotopes_dict = {
                                        'm/z': hill_mz_2,
                                        'intensity': hills_dict['hills_intensity_apex'][idx_2],
                                    }

                                    tmp_candidates.append(local_isotopes_dict)

                if len(tmp_candidates):
                    outmgf.write('BEGIN IONS\n')
                    outmgf.write('TITLE=%s.%d.%d.%d\n' % (basename_mzml, t_i, t_i, ms1_ch))
                    outmgf.write('RTINSECONDS=%f\n' % (ms1_RT * 60, ))
                    outmgf.write('PEPMASS=%f %f\n' % (ms1_mz, ms1_I))
                    outmgf.write('CHARGE=%d+\n' % (ms1_ch, ))
                    outmgf.write('MS1Intensity=%f\n' % (ms1_I, ))
                    for local_fragment in tmp_candidates:
                        outmgf.write('%f %f\n' % (local_fragment['m/z'], local_fragment['intensity']))
                    outmgf.write('END IONS\n\n')
                    t_i += 1


        logger.info('Chunk ready')


from . import main, main_dia
import argparse
from copy import deepcopy
import logging

def run():
    parser = argparse.ArgumentParser(
        description='A feature detection LC-MS1 spectra',
        epilog='''

    Example usage
    -------------
    $ biosaur2 input.mzML
    -------------
    ''',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('files', help='input mzML files', nargs='+')
    parser.add_argument('-mini', help='min intensity', default=1, type=float)
    parser.add_argument('-minmz', help='min mz', default=350, type=float)
    parser.add_argument('-maxmz', help='max mz', default=1500, type=float)
    parser.add_argument('-pasefmini', help='min intensity after combining hills in PASEF analysis', default=100, type=float)
    parser.add_argument('-htol', help='mass accuracy for hills in ppm', default=8, type=float)
    parser.add_argument('-itol', help='mass accuracy for isotopes in ppm', default=8, type=float)
    parser.add_argument('-paseftol', help='ion mobility accuracy for hills', default=0.05, type=float)
    parser.add_argument('-nm', help='negative mode. 1-true, 0-false', default=0, type=int)
    parser.add_argument('-o', help='path to output features file', default='')
    parser.add_argument('-hvf', help='hill valley factor', default=1.3, type=float)
    # parser.add_argument('-fdr', help='protein fdr filter in %%', default=1.0, type=float)
    # parser.add_argument('-i', help='minimum number of isotopes', default=2, type=int)
    parser.add_argument('-minlh', help='minimum length for hill', default=2, type=int)
    parser.add_argument('-pasefminlh', help='minimum length for pasef hill', default=1, type=int)
    # parser.add_argument('-ts', help='Two-stage RT training: 0 - turn off, 1 - turn one, 2 - turn on and use additive model in the first stage (Default)', default=2, type=int)
    # parser.add_argument('-sc', help='minimum number of scans for peptide feature', default=3, type=int)
    # parser.add_argument('-lmin', help='min length of peptides', default=7, type=int)
    # parser.add_argument('-lmax', help='max length of peptides', default=30, type=int)
    # parser.add_argument('-e', help='cleavage rule in quotes!. X!Tandem style for cleavage rules', default='[RK]|{P}')
    # parser.add_argument('-mc', help='number of missed cleavages', default=0, type=int)
    parser.add_argument('-cmin', help='min charge', default=1, type=int)
    parser.add_argument('-cmax', help='max charge', default=6, type=int)
    parser.add_argument('-nprocs', help='number of processes', default=4, type=int)
    # parser.add_argument('-fmods', help='fixed modifications. in mass1@aminoacid1,mass2@aminoacid2 format', default='57.021464@C')
    # parser.add_argument('-ad', help='add decoy', default=0, type=int)
    # parser.add_argument('-ml', help='use machine learning for PFMs', default=1, type=int)
    # parser.add_argument('-prefix', help='decoy prefix', default='DECOY_')
    # parser.add_argument('-nproc',   help='number of processes', default=1, type=int)
    # parser.add_argument('-elude', help='path to elude binary file. If empty, the built-in additive model will be used for RT prediction', default='')
    # parser.add_argument('-deeplc', help='path to deeplc', default='')
    # parser.add_argument('-deeplc_library', help='path to deeplc library', default='')
    # parser.add_argument('-pl', help='path to list of peptides for RT calibration', default='')
    parser.add_argument('-dia',  help='create mgf file for DIA MS/MS', action='store_true')
    parser.add_argument('-diahtol', help='mass accuracy for DIA hills in ppm', default=25, type=float)
    parser.add_argument('-diaminlh', help='minimum length for dia hill', default=1, type=int)
    parser.add_argument('-mgf', help='path to output mgf file', default='')
    parser.add_argument('-debug', help='log debugging information', action='store_true')
    parser.add_argument('-tof', help='smart tof processing', action='store_true')
    parser.add_argument('-profile', help='profile processing', action='store_true')
    # parser.add_argument('-diaitol', help='mass accuracy for DIA isotopes in ppm', default=25, type=float)
    args = vars(parser.parse_args())
    logging.basicConfig(format='%(levelname)9s: %(asctime)s %(message)s',
            datefmt='[%H:%M:%S]', level=[logging.INFO, logging.DEBUG][args['debug']])
    logging.getLogger('matplotlib').setLevel(logging.WARNING)
    logger = logging.getLogger(__name__)
    logger.debug('Starting with args: %s', args)

    for filename in args['files']:
        logger.info('Starting file: %s', filename)
        if 1:
            args['file'] = filename
            main.process_file(deepcopy(args))
            logger.info('Feature detection is finished for file: %s', filename)
            if args['dia']:
                main_dia.process_file(deepcopy(args))
        
        # except Exception as e:
        #     logger.error(e)
        #     logger.error('Feature detection failed for file: %s', filename)

if __name__ == '__main__':
    run()



cimport cython
from cython.parallel cimport prange, parallel
cimport openmp
import numpy as np
cimport numpy as np
import itertools
import math
from collections import Counter, defaultdict
from copy import copy

np.import_array()

cdef dict charge_ban_map

charge_ban_map = {
    8: (1, 2, 4, ),
    7: (1, ),
    6: (1, 2, 3, ),
    5: (1, ),
    4: (1, 2, ),
    3: (1, ),
    2: (1, ),
    1: (1, ),
}


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(True)
def get_and_calc_values_for_cos_corr(dict hills_dict, int idx_1):

    hill_idict_1 = hills_dict['hills_idict'][idx_1]
    if hill_idict_1 is None:
        hill_idict_1 = dict()
        for scan_id_val, intensity_val in zip(hills_dict['hills_scan_lists'][idx_1], hills_dict['hills_intensity_array'][idx_1]):
            hill_idict_1[scan_id_val] = intensity_val
        hills_dict['hills_idict'][idx_1] = hill_idict_1

    hill_sqrt_of_i_1 = hills_dict['hill_sqrt_of_i'][idx_1]
    if hill_sqrt_of_i_1 is None:
        hill_sqrt_of_i_1 = math.sqrt(sum(v**2 for v in hill_idict_1.values()))
        hills_dict['hill_sqrt_of_i'][idx_1] = hill_sqrt_of_i_1

    return hills_dict, hill_idict_1, hill_sqrt_of_i_1


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(True)
def get_and_calc_apex_intensity_and_scan(dict hills_dict, int idx_1):

    hill_intensity_apex_1 = hills_dict['hills_intensity_apex'][idx_1]
    hill_scan_apex_1 = hills_dict['hills_scan_apex'][idx_1]
    if hill_intensity_apex_1 is None:

        hill_intensity_apex_1 = 0

        for int_val, scan_val in zip(hills_dict['hills_intensity_array'][idx_1], hills_dict['hills_scan_lists'][idx_1]):
            if int_val > hill_intensity_apex_1:
                hill_intensity_apex_1 = int_val
                hill_scan_apex_1 = scan_val

        hills_dict['hills_intensity_apex'][idx_1] = hill_intensity_apex_1
        hills_dict['hills_scan_apex'][idx_1] = hill_scan_apex_1

    return hills_dict, hill_intensity_apex_1, hill_scan_apex_1


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(True)
def cos_correlation(set hill_scans_1, dict hill_idict_1, float hill_sqrt_of_i_1, set hill_scans_2, dict hill_idict_2, float hill_sqrt_of_i_2):

    cdef int i
    cdef float top

    top = 0
    for i in  hill_scans_1.intersection(hill_scans_2):
        top += hill_idict_1.get(i, 0) * hill_idict_2.get(i, 0)
    
    return top / (hill_sqrt_of_i_1 * hill_sqrt_of_i_2)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(True)
def get_initial_isotopes(dict hills_dict, float isotopes_mass_accuracy, list isotopes_list, dict a, int min_charge, int max_charge, float mz_step, bint paseftol, float faims_val, list sorted_idx_child_process):

    cdef list ready, charges, hill_scans_1_list, hill_scans_2_list, candidates, tmp_candidates
    cdef int idx_1, hill_idx_1, idx_2, hill_idx_2, im_to_check_fast, hill_scans_1_list_first, hill_scans_1_list_last
    cdef int hill_scans_2_list_first, hill_scans_2_list_last, charge, isotope_number, m_to_check_fast
    cdef float im_mz_1, hill_mz_1, im_mz_2, hill_mz_2, m_to_check, mass_diff_abs, cos_cor_RT
    cdef float hill_sqrt_of_i_1, hill_sqrt_of_i_2
    cdef dict banned_charges, hill_idict_1, hill_idict_2, local_isotopes_dict
    cdef set hill_scans_1, hill_scans_2

    ready = []

    charges = list(range(min_charge, max_charge + 1, 1)[::-1])

    for idx_1 in sorted_idx_child_process:
        hill_idx_1 = hills_dict['hills_idx_array_unique'][idx_1]
        hill_mz_1 = hills_dict['hills_mz_median'][idx_1]

    #for (idx_1, hill_idx_1), hill_mz_1 in sorted(list(zip(list(enumerate(hills_dict['hills_idx_array_unique'])), hills_dict['hills_mz_median'])), key=lambda x: x[-1]):

        if paseftol is not False:
            im_mz_1 = hills_dict['hills_im_median'][idx_1]
            im_to_check_fast = int(im_mz_1 / paseftol)

        banned_charges = dict()

        hill_scans_1 = hills_dict['hills_scan_sets'][idx_1]
        hill_scans_1_list = hills_dict['hills_scan_lists'][idx_1]
        hill_scans_1_list_first, hill_scans_1_list_last = hill_scans_1_list[0], hill_scans_1_list[-1]

        mz_tol = isotopes_mass_accuracy * 1e-6 * hill_mz_1

        for charge in charges:
            
            candidates = []

            for isotope_number in isotopes_list[1:]:

                tmp_candidates = []

                m_to_check = hill_mz_1 + (1.00335 * isotope_number / charge)
                m_to_check_fast = int(m_to_check/mz_step)

                for idx_2, hill_scans_2_list_first, hill_scans_2_list_last in hills_dict['hills_mz_median_fast_dict'][m_to_check_fast]:

                    if paseftol is False or idx_2 in hills_dict['hills_im_median_fast_dict'][im_to_check_fast]:

                        if not hill_scans_1_list_last < hill_scans_2_list_first and not hill_scans_2_list_last < hill_scans_1_list_first:
                            hill_mz_2 = hills_dict['hills_mz_median'][idx_2]
                            mass_diff_abs = hill_mz_2 - m_to_check

                            if abs(mass_diff_abs) <= mz_tol:
                                hill_scans_2 = hills_dict['hills_scan_sets'][idx_2]
                                if len(hill_scans_1.intersection(hill_scans_2)) >= 1:


                                    hills_dict, hill_idict_1, hill_sqrt_of_i_1 = get_and_calc_values_for_cos_corr(hills_dict, idx_1)
                                    hills_dict, hill_idict_2, hill_sqrt_of_i_2 = get_and_calc_values_for_cos_corr(hills_dict, idx_2)

                                    cos_cor_RT = cos_correlation(hill_scans_1, hill_idict_1, hill_sqrt_of_i_1, hill_scans_2, hill_idict_2, hill_sqrt_of_i_2)

                                    if cos_cor_RT >= 0.6:

                                        hill_idx_2 = hills_dict['hills_idx_array_unique'][idx_2]

                                        hills_dict, _, _ = get_and_calc_apex_intensity_and_scan(hills_dict, idx_1)
                                        hills_dict, _, _ = get_and_calc_apex_intensity_and_scan(hills_dict, idx_2)

                                        local_isotopes_dict = {
                                            'isotope_number': isotope_number,
                                            'isotope_hill_idx': hill_idx_2,
                                            'isotope_idx': idx_2,
                                            'cos_cor': cos_cor_RT,
                                            'mass_diff_ppm': mass_diff_abs/m_to_check*1e6,
                                        }

                                        tmp_candidates.append(local_isotopes_dict)



                if len(tmp_candidates):
                    candidates.append(tmp_candidates)

                if len(candidates) < isotope_number:
                    break



            if len(candidates) >= banned_charges.get(charge, 1):

                neutral_mass = hill_mz_1 * charge

                tmp_intensity = a[int(100 * (neutral_mass // 100))]

                _, hill_intensity_apex_1, hill_scan_apex_1 = get_and_calc_apex_intensity_and_scan(hills_dict, idx_1)
                mono_hills_scan_lists =  hills_dict['hills_scan_lists'][idx_1]
                mono_hills_intensity_list =  hills_dict['hills_intensity_array'][idx_1]

                all_theoretical_int = [
                    hill_intensity_apex_1 *
                    tmp_intensity[z] /
                    tmp_intensity[0] for z in isotopes_list]

                for iter_candidates in itertools.product(*candidates):

                    all_exp_intensity = [hill_intensity_apex_1, ]

                    for local_isotopes_dict in iter_candidates:

                        idx_2 = local_isotopes_dict['isotope_idx']

                        _, hill_intensity_apex_2, hill_scan_apex_2 = get_and_calc_apex_intensity_and_scan(hills_dict, idx_2)
                        
                        all_exp_intensity.append(hill_intensity_apex_2)
                    cos_corr, number_of_passed_isotopes = checking_cos_correlation_for_carbon(all_theoretical_int, all_exp_intensity, 0.6)

                    if cos_corr:

                        iter_candidates = iter_candidates[:number_of_passed_isotopes]

                        local_res_dict = {
                            'monoisotope hill idx': hill_idx_1,
                            'monoisotope idx': idx_1,
                            'cos_cor_isotopes': cos_corr,
                            'hill_mz_1': hill_mz_1,
                            'isotopes': iter_candidates,
                            'nIsotopes': number_of_passed_isotopes,
                            'charge': charge,
                            'FAIMS': faims_val,
                            'im': 0 if paseftol is False else im_mz_1,
                            'intensity_array_for_cos_corr': [all_theoretical_int[:number_of_passed_isotopes+1], all_exp_intensity[:number_of_passed_isotopes+1]],
                        }

                        ready.append(local_res_dict)

                        for ch_v in charge_ban_map[charge]:
                            banned_charges[ch_v] = max(number_of_passed_isotopes, banned_charges.get(ch_v, 1))
    return ready

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(True)
def cos_correlation_new(list theoretical_list, list experimental_list):
    cdef float theor_total_sum, top, i1, i2, bottom, numb, averagineExplained
    cdef int suit_len


    theor_total_sum = sum(theoretical_list)
    suit_len = min(len(theoretical_list), len(experimental_list))
    theoretical_list = theoretical_list[:suit_len]
    experimental_list = experimental_list[:suit_len]

    top = 0

    for i1, i2 in zip(theoretical_list, experimental_list):
        top += i1 * i2

    if not top:
        return 0, 0
    else:

        bottom = math.sqrt(sum([numb * numb for numb in theoretical_list])) * \
            math.sqrt(sum([numb * numb for numb in experimental_list]))

        averagineExplained = sum(theoretical_list) / theor_total_sum

        return top / bottom, averagineExplained

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(True)
def checking_cos_correlation_for_carbon(list theoretical_list, list experimental_list, float thresh):

    cdef float averagineCorrelation, averagineExplained, best_cor
    cdef int best_pos, pos

    best_pos = 1
    best_cor = 0

    pos = len(experimental_list)

    while pos != 1:

        averagineCorrelation, averagineExplained = cos_correlation_new(theoretical_list, experimental_list[:pos])

        if averagineExplained >= 0.5 and averagineCorrelation >= thresh:
            if averagineCorrelation > best_cor:
                best_cor = averagineCorrelation
                best_pos = pos

            break

        pos -= 1

    return best_cor, best_pos


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(True)
def get_fast_dict(np.ndarray mz_sorted, float mz_step, list basic_id_sorted):

    cdef dict fast_dict
    cdef list fast_array
    cdef int idx, fm

    fast_dict = dict()
    fast_array = list((mz_sorted/mz_step).astype(int))
    for idx, fm in zip(basic_id_sorted, fast_array):
        if fm not in fast_dict:
            fast_dict[fm] = [idx, ]
        else:
            fast_dict[fm].append(idx)
    return fast_dict, fast_array


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(True)
def meanfilt(list data, int window_width):
    cdef np.ndarray cumsum_vec, ma_vec_array
    cdef list ma_vec
    cumsum_vec = np.cumsum(np.insert(data, 0, 0))
    ma_vec_array = (cumsum_vec[window_width:] -
              cumsum_vec[:-window_width]) / window_width
    ma_vec = data[:1] + list(ma_vec_array[1:]) + data[-1:]
    return ma_vec


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(True)
def split_peaks(dict hills_dict, list data_for_analyse_tmp, dict args, dict counter_hills_idx, list sorted_idx_child_process, np.ndarray sorted_idx_array_child_process, int nproc, int checked_id):

    cdef float hillValleyFactor, min_val, mult_val
    cdef int min_length_hill, idx_start, idx_end, cur_new_idx, idx_1, hill_idx, hill_length, c_len, l_idx
    cdef np.ndarray idx_sort, tmp_scans, tmp_orig_idx, new_index_list
    cdef list tmp_intensity, smothed_intensity, min_idx_list, recheck_r_r 

    new_index_list = copy(sorted_idx_array_child_process)

    hillValleyFactor = args['hvf']
    min_length_hill = args['minlh']
    min_length_hill = max(2, min_length_hill)

    idx_start = 0
    idx_end = 0

    cur_new_idx = checked_id + max(sorted_idx_child_process) + 1

    for hill_idx in sorted_idx_child_process:

        hill_length = counter_hills_idx[hill_idx]
        idx_end = idx_start + hill_length


        if hill_length >= min_length_hill * 2:


            tmp_scans = hills_dict['scan_idx_array'][checked_id+idx_start:checked_id+idx_end]
            tmp_orig_idx = hills_dict['orig_idx_array'][checked_id+idx_start:checked_id+idx_end]
            tmp_intensity = [data_for_analyse_tmp[scan_val][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]

            smothed_intensity = meanfilt(tmp_intensity, 2)
            c_len = hill_length - min_length_hill
            idx = int(min_length_hill) - 1
            min_idx_list = []
            min_val = 0
            l_idx = 0
            recheck_r_r = []


            while idx <= c_len:

                if len(min_idx_list) and idx >= min_idx_list[-1] + min_length_hill:
                    l_idx = min_idx_list[-1]

                l_r = max(smothed_intensity[l_idx:idx]) / float(smothed_intensity[idx])
                if l_r >= hillValleyFactor:
                    r_r = max(smothed_intensity[idx+1:]) / float(smothed_intensity[idx])
                    if r_r >= hillValleyFactor:
                        mult_val = l_r * r_r
                        include_factor = (1 if l_r > r_r else 0)
                        if (min_length_hill <= idx + include_factor <= c_len):
                            if not len(min_idx_list) or idx + include_factor >= min_idx_list[-1] + min_length_hill:
                                min_idx_list.append(idx + include_factor)
                                recheck_r_r.append(idx)
                                min_val = mult_val
                            elif mult_val > min_val:
                                min_idx_list[-1] = idx + include_factor
                                recheck_r_r[-1] = idx
                                min_val = mult_val
                idx += 1
            if len(min_idx_list):
                for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):
                    r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])
                    if r_r >= hillValleyFactor:
                        new_index_list[idx_start+min_idx:idx_start+hill_length] = cur_new_idx
                        cur_new_idx += 1

        idx_start = idx_end

    return new_index_list




@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(True)
def split_peaks_old(dict hills_dict, list data_for_analyse_tmp, dict args):

    cdef float hillValleyFactor, min_val, mult_val
    cdef int min_length_hill, idx_start, idx_end, cur_new_idx, idx_1, hill_idx, hill_length, c_len, l_idx
    cdef np.ndarray tmp_hill_length, idx_minl, idx_sort, tmp_scans, tmp_orig_idx
    cdef list tmp_intensity, smothed_intensity, min_idx_list, recheck_r_r

    hillValleyFactor = args['hvf']
    min_length_hill = args['minlh']
    min_length_hill = max(2, min_length_hill)


    hills_dict['orig_idx_array'] = np.array(hills_dict['orig_idx_array'])
    hills_dict['scan_idx_array'] = np.array(hills_dict['scan_idx_array'])
    hills_dict['hills_idx_array'] = np.array(hills_dict['hills_idx_array'])

    counter_hills_idx = Counter(hills_dict['hills_idx_array'])

    tmp_hill_length = np.array([counter_hills_idx[hill_idx] for hill_idx in hills_dict['hills_idx_array']])
    idx_minl = tmp_hill_length >= min_length_hill
    hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_minl]
    hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_minl]
    hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_minl]

    if len(hills_dict['orig_idx_array']):

        idx_sort = np.argsort(hills_dict['hills_idx_array'] + (hills_dict['scan_idx_array'] / (hills_dict['scan_idx_array'].max()+1)))
        hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_sort]
        hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_sort]
        hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_sort]

        hills_dict['hills_idx_array_unique'] = sorted(list(set(hills_dict['hills_idx_array'])))

        idx_start = 0
        idx_end = 0

        cur_new_idx = max(hills_dict['hills_idx_array_unique']) + 1

        for idx_1, hill_idx in enumerate(hills_dict['hills_idx_array_unique']):

            hill_length = counter_hills_idx[hill_idx]
            idx_end = idx_start + hill_length

            if hill_length >= min_length_hill * 2:


                tmp_scans = hills_dict['scan_idx_array'][idx_start:idx_end]
                tmp_orig_idx = hills_dict['orig_idx_array'][idx_start:idx_end]
                tmp_intensity = [data_for_analyse_tmp[scan_val]['intensity array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]

                smothed_intensity = meanfilt(tmp_intensity, 2)
                c_len = hill_length - min_length_hill
                idx = int(min_length_hill) - 1
                min_idx_list = []
                min_val = 0
                l_idx = 0
                recheck_r_r = []

                while idx <= c_len:

                    if len(min_idx_list) and idx >= min_idx_list[-1] + min_length_hill:
                        l_idx = min_idx_list[-1]

                    l_r = max(smothed_intensity[l_idx:idx]) / float(smothed_intensity[idx])
                    if l_r >= hillValleyFactor:
                        r_r = max(smothed_intensity[idx+1:]) / float(smothed_intensity[idx])
                        if r_r >= hillValleyFactor:
                            mult_val = l_r * r_r
                            include_factor = (1 if l_r > r_r else 0)
                            if (min_length_hill <= idx + include_factor <= c_len):
                                if not len(min_idx_list) or idx + include_factor >= min_idx_list[-1] + min_length_hill:
                                    min_idx_list.append(idx + include_factor)
                                    recheck_r_r.append(idx)
                                    min_val = mult_val
                                elif mult_val > min_val:
                                    min_idx_list[-1] = idx + include_factor
                                    recheck_r_r[-1] = idx
                                    min_val = mult_val
                    idx += 1
                if len(min_idx_list):
                    for min_idx, end_idx, recheck_idx in zip(min_idx_list, min_idx_list[1:] + [idx_start+hill_length, ], recheck_r_r):
                        r_r = max(smothed_intensity[recheck_idx+1:end_idx]) / float(smothed_intensity[recheck_idx])
                        if r_r >= hillValleyFactor:
                            hills_dict['hills_idx_array'][idx_start+min_idx:idx_start+hill_length] = cur_new_idx
                            cur_new_idx += 1

            idx_start = idx_end

        hills_dict['hills_idx_array'] = list(hills_dict['hills_idx_array'])
        del hills_dict['hills_idx_array_unique']

    return hills_dict



@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(True)
def detect_hills(list data_for_analyse_tmp, dict args, float mz_step, bint paseftol, bint dia=False):

    cdef dict hills_dict, prev_fast_dict, z
    cdef list total_num_hills, total_mass_diff, spec_mean_mass_accuracy, basic_id_sorted, all_idx, all_prevs
    cdef set banned_prev_idx_set
    cdef float hill_mass_accuracy, best_intensity, mz_cur, cur_intensity, cur_mass_diff_with_sign, cur_mass_diff, best_mass_diff
    cdef int last_idx, prev_idx, spec_idx, len_mz, idx, fm, fi, best_idx_prev, idx_prev
    cdef np.ndarray idx_for_sort, mz_sorted, im_sorted
    cdef bint flag1, flag2, flag3

    hills_dict = {}

    if dia is False:
        hill_mass_accuracy = args['htol']
    else:
        hill_mass_accuracy = args['diahtol']

    hills_dict['hills_idx_array'] = []
    hills_dict['orig_idx_array'] = []
    hills_dict['scan_idx_array'] = []
    hills_dict['mzs_array'] = []
    hills_dict['intensity_array'] = []
    if paseftol is not False:
        hills_dict['im_array'] = []
        prev_fast_dict_im = dict()

    last_idx = -1
    prev_idx = -1
    prev_fast_dict = dict()

    total_num_hills = []
    total_mass_diff = []

    for spec_idx, z in enumerate(data_for_analyse_tmp):

        len_mz = len(z['m/z array'])

        hills_dict['hills_idx_array'].extend(list(range(last_idx+1, last_idx+1+len_mz, 1)))
        hills_dict['orig_idx_array'].extend(range(len_mz))
        hills_dict['scan_idx_array'].extend([spec_idx] * len_mz)

        idx_for_sort = np.argsort(z['intensity array'])[::-1]

        mz_sorted = z['m/z array'][idx_for_sort]
        basic_id_sorted = list(np.array(range(len_mz))[idx_for_sort])

        hills_dict['mzs_array'].extend(z['m/z array'])
        hills_dict['intensity_array'].extend(z['intensity array'])

        if paseftol is not False:
            im_sorted = z['mean inverse reduced ion mobility array'][idx_for_sort]
            hills_dict['im_array'].extend(z['mean inverse reduced ion mobility array'])

            fast_dict_im, fast_array_im = get_fast_dict(im_sorted, paseftol, basic_id_sorted)

        fast_dict, fast_array = get_fast_dict(mz_sorted, mz_step, basic_id_sorted)

        banned_prev_idx_set = set()

        for idx, fm, fi in zip(basic_id_sorted, fast_array, (fast_array if paseftol is False else fast_array_im)):

            flag1 = fm in prev_fast_dict
            flag2 = fm-1 in prev_fast_dict
            flag3 = fm+1 in prev_fast_dict

            if flag1 or flag2 or flag3:

                if flag1:
                    all_idx = prev_fast_dict[fm]
                    if flag2:
                        all_idx += prev_fast_dict[fm-1]
                    if flag3:
                        all_idx += prev_fast_dict[fm+1]
                elif flag2:
                    all_idx = prev_fast_dict[fm-1]
                    if flag3:
                        all_idx += prev_fast_dict[fm+1]
                elif flag3:
                    all_idx = prev_fast_dict[fm+1]

                best_intensity = 0
                best_idx_prev = 0
                mz_cur = z['m/z array'][idx]


                all_prevs = [[idx_prev, data_for_analyse_tmp[spec_idx-1]['intensity array'][idx_prev]] for idx_prev in all_idx if idx_prev not in banned_prev_idx_set and (paseftol is False or idx_prev in prev_fast_dict_im[fi])]
                for idx_prev, cur_intensity in all_prevs:
                    cur_mass_diff_with_sign = (mz_cur - data_for_analyse_tmp[spec_idx-1]['m/z array'][idx_prev]) / mz_cur * 1e6
                    cur_mass_diff = abs(cur_mass_diff_with_sign)
                    if cur_mass_diff <= hill_mass_accuracy and cur_intensity >= best_intensity:
                        best_mass_diff = cur_mass_diff
                        best_intensity = cur_intensity
                        best_idx_prev = idx_prev
                        hills_dict['hills_idx_array'][last_idx+1+idx] = hills_dict['hills_idx_array'][prev_idx+1+idx_prev]
                        
                if best_idx_prev != 0:
                    banned_prev_idx_set.add(best_idx_prev)


        prev_fast_dict = fast_dict
        if paseftol is not False:
            prev_fast_dict_im = fast_dict_im
        prev_idx = last_idx
        last_idx = last_idx+len_mz

    return hills_dict


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(True)
def process_hills(dict hills_dict, list data_for_analyse_tmp, float mz_step, bint paseftol, dict args, bint dia=False):

    cdef dict 
    cdef float mz_median, i_sum_tmp, mz_val_tmp, i_val_tmp, im_median
    cdef list tmp_intensity, tmp_mz_array, tmp_im_array, tmp_scans_list, tmp_scans, tmp_orig_idx
    cdef set tmp_scans_set
    cdef tuple tmp_val
    cdef int min_length_hill, idx_start, idx_end, idx_1, hill_idx, hill_length, mz_median_int, im_median_int
    cdef np.ndarray tmp_hill_length, idx_minl


    counter_hills_idx = Counter(hills_dict['hills_idx_array'])
    if dia is False:
        min_length_hill = args['minlh']
    else:
        min_length_hill = args['diaminlh']

    hills_dict['orig_idx_array'] = np.array(hills_dict['orig_idx_array'])
    hills_dict['scan_idx_array'] = np.array(hills_dict['scan_idx_array'])
    hills_dict['hills_idx_array'] = np.array(hills_dict['hills_idx_array'])


    tmp_hill_length = np.array([counter_hills_idx[hill_idx] for hill_idx in hills_dict['hills_idx_array']])
    idx_minl = tmp_hill_length >= min_length_hill
    hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_minl]
    hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_minl]
    hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_minl]

    if len(hills_dict['hills_idx_array']):

        idx_sort = np.argsort(hills_dict['hills_idx_array'] + (hills_dict['scan_idx_array'] / (hills_dict['scan_idx_array'].max()+1)))
        hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_sort]
        hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_sort]
        hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_sort]

        hills_dict['hills_idx_array_unique'] = sorted(list(set(hills_dict['hills_idx_array'])))
        hills_dict['hills_mz_median'] = []
        hills_dict['hills_mz_median_fast_dict'] = defaultdict(list)
        if paseftol is not False:
            hills_dict['hills_im_median'] = []
            hills_dict['hills_im_median_fast_dict'] = defaultdict(set)

        hills_dict['hills_intensity_array'] = []
        hills_dict['hills_scan_sets'] = []
        hills_dict['hills_scan_lists'] = []
        hills_dict['hills_lengths'] = []
        hills_dict['tmp_mz_array'] = []

        hills_dict['scan_idx_array'] = list(hills_dict['scan_idx_array'])
        hills_dict['orig_idx_array'] = list(hills_dict['orig_idx_array'])
        hills_dict['hills_idx_array'] = list(hills_dict['hills_idx_array'])

        idx_start = 0
        idx_end = 0


        for idx_1, hill_idx in enumerate(hills_dict['hills_idx_array_unique']):

            hill_length = counter_hills_idx[hill_idx]
            idx_end = idx_start + hill_length

            tmp_scans = hills_dict['scan_idx_array'][idx_start:idx_end]
            tmp_orig_idx = hills_dict['orig_idx_array'][idx_start:idx_end]

            tmp_intensity = [data_for_analyse_tmp[scan_val]['intensity array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
            tmp_mz_array = [data_for_analyse_tmp[scan_val]['m/z array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
            mz_median = 0
            i_sum_tmp = 0
            for mz_val_tmp, i_val_tmp in zip(tmp_mz_array, tmp_intensity):
                mz_median += mz_val_tmp * i_val_tmp
                i_sum_tmp += i_val_tmp
            mz_median = mz_median / i_sum_tmp
            if paseftol is not False:
                tmp_im_array = [data_for_analyse_tmp[scan_val]['mean inverse reduced ion mobility array'][orig_idx_val] for orig_idx_val, scan_val in zip(tmp_orig_idx, tmp_scans)]
                im_median = np.average(tmp_im_array, weights=tmp_intensity)
            tmp_scans_list = tmp_scans
            tmp_scans_set = set(tmp_scans)

            idx_start = idx_end

            hills_dict['hills_mz_median'].append(mz_median)

            mz_median_int = int(mz_median/mz_step)
            tmp_val = (idx_1, tmp_scans_list[0], tmp_scans_list[-1])
            hills_dict['hills_mz_median_fast_dict'][mz_median_int-1].append(tmp_val)
            hills_dict['hills_mz_median_fast_dict'][mz_median_int].append(tmp_val)
            hills_dict['hills_mz_median_fast_dict'][mz_median_int+1].append(tmp_val)

            if paseftol is not False:
                hills_dict['hills_im_median'].append(im_median)

                im_median_int = int(im_median/paseftol)
                hills_dict['hills_im_median_fast_dict'][im_median_int-1].add(idx_1)
                hills_dict['hills_im_median_fast_dict'][im_median_int].add(idx_1)
                hills_dict['hills_im_median_fast_dict'][im_median_int+1].add(idx_1)


            hills_dict['hills_intensity_array'].append(tmp_intensity)
            hills_dict['hills_scan_sets'].append(tmp_scans_set)
            hills_dict['hills_scan_lists'].append(tmp_scans_list)
            hills_dict['hills_lengths'].append(hill_length)
            hills_dict['tmp_mz_array'].append(tmp_mz_array)


        hills_dict['hills_idict'] = [None] * len(hills_dict['hills_idx_array_unique'])
        hills_dict['hill_sqrt_of_i'] = [None] * len(hills_dict['hills_idx_array_unique'])
        hills_dict['hills_intensity_apex'] = [None] * len(hills_dict['hills_idx_array_unique'])
        hills_dict['hills_scan_apex'] = [None] * len(hills_dict['hills_idx_array_unique'])

    return hills_dict


from pyteomics import mzml
import numpy as np
from collections import defaultdict, Counter
from os import path
import math
from scipy.optimize import curve_fit
import logging
logger = logging.getLogger(__name__)
from .cutils import get_fast_dict, get_and_calc_apex_intensity_and_scan

class MS1OnlyMzML(mzml.MzML): 
     _default_iter_path = '//spectrum[./*[local-name()="cvParam" and @name="ms level" and @value="1"]]' 
     _use_index = False 
     _iterative = False

def noisygaus(x, a, x0, sigma, b):
    return a * np.exp(-(x - x0) ** 2 / (2 * sigma ** 2)) + b

def calibrate_mass(bwidth, mass_left, mass_right, true_md):

    bbins = np.arange(-mass_left, mass_right, bwidth)
    H1, b1 = np.histogram(true_md, bins=bbins)
    b1 = b1 + bwidth
    b1 = b1[:-1]

    popt, pcov = curve_fit(noisygaus, b1, H1, p0=[1, np.median(true_md), 1, 1])
    mass_shift, mass_sigma = popt[1], abs(popt[2])
    return mass_shift, mass_sigma, pcov[0][0]


def calc_peptide_features(hills_dict, peptide_features, negative_mode, faims_val, RT_dict, data_start_id):

    for pep_feature in peptide_features:

        pep_feature['mz'] = pep_feature['hill_mz_1']
        pep_feature['isoerror'] = pep_feature['isotopes'][0]['mass_diff_ppm']
        pep_feature['isoerror2'] = pep_feature['isotopes'][1]['mass_diff_ppm'] if len(pep_feature['isotopes']) > 1 else -100
        pep_feature['nScans'] = hills_dict['hills_lengths'][pep_feature['monoisotope idx']]

        pep_feature['massCalib'] = pep_feature['mz'] * pep_feature['charge'] - 1.0072765 * pep_feature['charge'] * (-1 if negative_mode else 1)

        hills_dict, _, _ = get_and_calc_apex_intensity_and_scan(hills_dict, pep_feature['monoisotope idx'])

        pep_feature['scanApex'] = hills_dict['hills_scan_apex'][pep_feature['monoisotope idx']]
        pep_feature['rtApex'] = RT_dict[hills_dict['hills_scan_apex'][pep_feature['monoisotope idx']]+data_start_id]
        pep_feature['intensityApex'] = hills_dict['hills_intensity_apex'][pep_feature['monoisotope idx']]
        pep_feature['rtStart'] = RT_dict[hills_dict['hills_scan_lists'][pep_feature['monoisotope idx']][0]+data_start_id]
        pep_feature['rtEnd'] = RT_dict[hills_dict['hills_scan_lists'][pep_feature['monoisotope idx']][-1]+data_start_id]
        pep_feature['mono_hills_scan_lists'] = hills_dict['hills_scan_lists'][pep_feature['monoisotope idx']]
        pep_feature['mono_hills_intensity_list'] =  hills_dict['hills_intensity_array'][pep_feature['monoisotope idx']]

    return peptide_features


def write_output(peptide_features, args, write_header=True):

    input_mzml_path = args['file']

    if args['o']:
        output_file = args['o']
    else:
        output_file = path.splitext(input_mzml_path)[0]\
            + path.extsep + 'features.tsv'

    columns_for_output = [
        'massCalib',
        'rtApex',
        'intensityApex',
        'charge',
        'nIsotopes',
        'nScans',
        'mz',
        'rtStart',
        'rtEnd',
        'FAIMS',
        'im',
        'mono_hills_scan_lists',
        'mono_hills_intensity_list',
        'scanApex',
        'isoerror2',
    ]

    if write_header:

        out_file = open(output_file, 'w')
        out_file.write('\t'.join(columns_for_output) + '\n')
        out_file.close()

    out_file = open(output_file, 'a')
    for pep_feature in peptide_features:
        out_file.write('\t'.join([str(pep_feature[col]) for col in columns_for_output]) + '\n')

    out_file.close()


def centroid_pasef_data(data_for_analyse_tmp, args, mz_step):

    cnt_ms1_scans = len(data_for_analyse_tmp)
    for spec_idx, z in enumerate(data_for_analyse_tmp):

        logger.info('PASEF scans analysis: %d/%d', spec_idx+1, cnt_ms1_scans)
        logger.info('number of m/z peaks in scan: %d', len(z['m/z array']))

        if 'ignore_ion_mobility' not in z:

            mz_ar_new = []
            intensity_ar_new = []
            ion_mobility_ar_new = []

            mz_ar = z['m/z array']
            intensity_ar = z['intensity array']
            ion_mobility_ar = z['mean inverse reduced ion mobility array']

            ion_mobility_accuracy = args['paseftol']
            ion_mobility_step = max(ion_mobility_ar) * ion_mobility_accuracy

            ion_mobility_ar_fast = (ion_mobility_ar/ion_mobility_step).astype(int)
            mz_ar_fast = (mz_ar/mz_step).astype(int)

            idx = np.argsort(mz_ar_fast)
            mz_ar_fast = mz_ar_fast[idx]
            ion_mobility_ar_fast = ion_mobility_ar_fast[idx]

            mz_ar = mz_ar[idx]
            intensity_ar = intensity_ar[idx]
            ion_mobility_ar = ion_mobility_ar[idx]

            max_peak_idx = len(mz_ar)

            banned_idx = set()

            peak_idx = 0
            while peak_idx < max_peak_idx:

                if peak_idx not in banned_idx:

                    mass_accuracy_cur = mz_ar[peak_idx] * 1e-6 * args['itol']

                    mz_val_int = mz_ar_fast[peak_idx]
                    ion_mob_val_int = ion_mobility_ar_fast[peak_idx]

                    tmp = [peak_idx, ]

                    peak_idx_2 = peak_idx + 1

                    while peak_idx_2 < max_peak_idx:


                        if peak_idx_2 not in banned_idx:

                            mz_val_int_2 = mz_ar_fast[peak_idx_2]
                            if mz_val_int_2 - mz_val_int > 1:
                                break
                            elif abs(mz_ar[peak_idx]-mz_ar[peak_idx_2]) <= mass_accuracy_cur:
                                ion_mob_val_int_2 = ion_mobility_ar_fast[peak_idx_2]
                                if abs(ion_mob_val_int - ion_mob_val_int_2) <= 1:
                                    tmp.append(peak_idx_2)
                                    peak_idx = peak_idx_2
                        peak_idx_2 += 1

                all_intensity = [intensity_ar[p_id] for p_id in tmp]
                i_val_new = sum(all_intensity)

                if i_val_new >= args['pasefmini'] and len(all_intensity) >= args['pasefminlh']:

                    all_mz = [mz_ar[p_id] for p_id in tmp]
                    all_ion_mob = [ion_mobility_ar[p_id] for p_id in tmp]

                    mz_val_new = np.average(all_mz, weights=all_intensity)
                    ion_mob_new = np.average(all_ion_mob, weights=all_intensity)

                    intensity_ar_new.append(i_val_new)
                    mz_ar_new.append(mz_val_new)
                    ion_mobility_ar_new.append(ion_mob_new)

                    banned_idx.update(tmp)

                peak_idx += 1

            data_for_analyse_tmp[spec_idx]['m/z array'] = np.array(mz_ar_new)
            data_for_analyse_tmp[spec_idx]['intensity array'] = np.array(intensity_ar_new)
            data_for_analyse_tmp[spec_idx]['mean inverse reduced ion mobility array'] = np.array(ion_mobility_ar_new)

        logger.info('number of m/z peaks in scan after centroiding: %d', len(data_for_analyse_tmp[spec_idx]['m/z array']))

    data_for_analyse_tmp = [z for z in data_for_analyse_tmp if len(z['m/z array'] > 0)]
    logger.info('Number of MS1 scans after combining ion mobility peaks: %d', len(data_for_analyse_tmp))

            # fast_dict = defaultdict(set)
            # for peak_idx, (mz_val_int, ion_mob_val_int) in enumerate(zip(mz_ar_fast, ion_mobility_ar_fast)):

            #     fast_dict[(mz_val_int-1, ion_mob_val_int)].add(peak_idx)
            #     fast_dict[(mz_val_int, ion_mob_val_int)].add(peak_idx)
            #     fast_dict[(mz_val_int+1, ion_mob_val_int)].add(peak_idx)

            #     fast_dict[(mz_val_int-1, ion_mob_val_int-1)].add(peak_idx)
            #     fast_dict[(mz_val_int, ion_mob_val_int-1)].add(peak_idx)
            #     fast_dict[(mz_val_int+1, ion_mob_val_int-1)].add(peak_idx)

            #     fast_dict[(mz_val_int-1, ion_mob_val_int+1)].add(peak_idx)
            #     fast_dict[(mz_val_int, ion_mob_val_int+1)].add(peak_idx)
            #     fast_dict[(mz_val_int+1, ion_mob_val_int+1)].add(peak_idx)


    #         print('HERE2')

    #         hill_length = []
    #         peak_idx_array = []
    #         for peak_idx, (mz_val_int, ion_mob_val_int) in enumerate(zip(mz_ar_fast, ion_mobility_ar_fast)):
    #             hill_length.append(len(fast_dict[(mz_val_int, ion_mob_val_int)]))
    #             peak_idx_array.append(peak_idx)
    #         peak_idx_array = np.array(peak_idx_array)


    #         print('HERE3')

    #         added_idx = set()
    #         idx_sort = np.argsort(hill_length)[::-1]
    #         for peak_idx in peak_idx_array[idx_sort]:
    #             if peak_idx not in added_idx:
    #                 mz_val_int = mz_ar_fast[peak_idx]
    #                 ion_mob_val_int = ion_mobility_ar_fast[peak_idx]
    #                 all_idx = set([p_id for p_id in fast_dict[(mz_val_int, ion_mob_val_int)] if p_id not in added_idx])
    #                 if len(all_idx):
    #                     added_idx.update(all_idx)

    #                     all_intensity = [intensity_ar[p_id] for p_id in all_idx]
    #                     i_val_new = sum(all_intensity)

    #                     if i_val_new >= args['pasefmini']:

    #                         all_mz = [mz_ar[p_id] for p_id in all_idx]
    #                         all_ion_mob = [ion_mobility_ar[p_id] for p_id in all_idx]

    #                         mz_val_new = np.average(all_mz, weights=all_intensity)
    #                         ion_mob_new = np.average(all_ion_mob, weights=all_intensity)

    #                         intensity_ar_new.append(i_val_new)
    #                         mz_ar_new.append(mz_val_new)
    #                         ion_mobility_ar_new.append(ion_mob_new)

    #         data_for_analyse_tmp[spec_idx]['m/z array'] = np.array(mz_ar_new)
    #         data_for_analyse_tmp[spec_idx]['intensity array'] = np.array(intensity_ar_new)
    #         data_for_analyse_tmp[spec_idx]['mean inverse reduced ion mobility array'] = np.array(ion_mobility_ar_new)

    # data_for_analyse_tmp = [z for z in data_for_analyse_tmp if len(z['m/z array'] > 0)]
    # print('Number of MS1 scans after combining ion mobility peaks: ', len(data_for_analyse_tmp))

    return data_for_analyse_tmp

def process_profile(data_for_analyse_tmp):

    data_for_analyse_tmp_out = []

    for z in data_for_analyse_tmp:

        best_mz = 0
        best_int = 0
        best_im = 0
        prev_mz = False
        prev_int = False

        threshold = 0.05

        ar1 = []
        ar2 = []
        ar3 = []
        for mzv, intv, imv in zip(z['m/z array'], z['intensity array'], z['mean inverse reduced ion mobility array']):
            if prev_mz is False:
                best_mz = mzv
                best_int = intv
                best_im = imv
            elif mzv - prev_mz > threshold:
                ar1.append(best_mz)
                ar2.append(best_int)
                ar3.append(best_im)
                best_mz = mzv
                best_int = intv
                best_im = imv
            elif best_int > prev_int and intv > prev_int:
                ar1.append(best_mz)
                ar2.append(best_int)
                ar3.append(best_im)
                best_mz = mzv
                best_int = intv
                best_im = imv
            elif intv > best_int:
                best_mz = mzv
                best_int = intv
                best_im = imv
            prev_mz = mzv
            prev_int = intv

        ar1.append(best_mz)
        ar2.append(best_int)
        ar3.append(best_im)

        z['m/z array'] = np.array(ar1)
        z['intensity array'] = np.array(ar2)
        z['mean inverse reduced ion mobility array'] = np.array(ar3)

        data_for_analyse_tmp_out.append(z)
    return data_for_analyse_tmp_out



def process_tof(data_for_analyse_tmp):

            # print(len(z['m/z array']))
    universal_dict = {}
    cnt = 0

    for z in data_for_analyse_tmp:

        fast_set = z['m/z array'] // 50
        if cnt == 1:


            for l in set(fast_set):
                idxt = fast_set == l
                true_i = np.log10(z['intensity array'])[idxt]

                if len(true_i) > 150:

                    i_left = true_i.min()
                    i_right = true_i.max()

                    i_shift, i_sigma, covvalue = calibrate_mass(0.05, i_left, i_right, true_i)
                    # median_val = 
                    print(i_shift, i_sigma, covvalue)
                    universal_dict[l] = 10**(i_shift + 3 * i_sigma)#10**(np.median(true_i[idxt]) * 2)
            

            
        thresholds = [universal_dict.get(zz, 150) for zz in list(fast_set)]
        idxt2 = z['intensity array'] <= thresholds
        z['intensity array'][idxt2] = -1


        idx = z['intensity array'] > 0
        z['intensity array'] = z['intensity array'][idx]
        z['m/z array'] = z['m/z array'][idx]
        z['mean inverse reduced ion mobility array'] = z['mean inverse reduced ion mobility array'][idx]



        cnt += 1

        data_for_analyse_tmp = [z for z in data_for_analyse_tmp if len(z['m/z array'])]

    return data_for_analyse_tmp


def process_mzml(args):

    input_mzml_path = args['file']
    min_intensity = args['mini']
    min_mz = args['minmz']
    max_mz = args['maxmz']

    skipped = 0
    data_for_analyse = []

    cnt = 0

    for z in MS1OnlyMzML(source=input_mzml_path):
        if z['ms level'] == 1:

            if 'mean inverse reduced ion mobility array' not in z:
                z['ignore_ion_mobility'] = True
                z['mean inverse reduced ion mobility array'] = np.zeros(len(z['m/z array']))

            idx = z['intensity array'] >= min_intensity
            z['intensity array'] = z['intensity array'][idx]
            z['m/z array'] = z['m/z array'][idx]
            z['mean inverse reduced ion mobility array'] = z['mean inverse reduced ion mobility array'][idx]

            idx = z['m/z array'] >= min_mz
            z['m/z array'] = z['m/z array'][idx]
            z['intensity array'] = z['intensity array'][idx]
            z['mean inverse reduced ion mobility array'] = z['mean inverse reduced ion mobility array'][idx]

            idx = z['m/z array'] <= max_mz
            z['m/z array'] = z['m/z array'][idx]
            z['intensity array'] = z['intensity array'][idx]
            z['mean inverse reduced ion mobility array'] = z['mean inverse reduced ion mobility array'][idx]

            idx = np.argsort(z['m/z array'])
            z['m/z array'] = z['m/z array'][idx]
            z['intensity array'] = z['intensity array'][idx]
            z['mean inverse reduced ion mobility array'] = z['mean inverse reduced ion mobility array'][idx]

            cnt += 1

            # if len(data_for_analyse) > 50:
            #     break

            if len(z['m/z array']):
                data_for_analyse.append(z)
            else:
                skipped += 1


    logger.info('Number of MS1 scans: %d', len(data_for_analyse))
    logger.info('Number of skipped MS1 scans: %d', skipped)

    if len(data_for_analyse) == 0:
        raise Exception('no MS1 scans in input file')

    return data_for_analyse



def process_mzml_dia(args):

    input_mzml_path = args['file']
    # min_intensity = args['mini']
    # min_mz = args['minmz']
    # max_mz = args['maxmz']
    min_intensity = 0
    min_mz = 1
    max_mz = 1e6

    skipped = 0
    data_for_analyse = []

    cnt = 0

    for z in mzml.read(input_mzml_path):
        if z['ms level'] == 2:

            if 'mean inverse reduced ion mobility array' not in z:
                z['ignore_ion_mobility'] = True
                z['mean inverse reduced ion mobility array'] = np.zeros(len(z['m/z array']))

            idx = z['intensity array'] >= min_intensity
            z['intensity array'] = z['intensity array'][idx]
            z['m/z array'] = z['m/z array'][idx]
            z['mean inverse reduced ion mobility array'] = z['mean inverse reduced ion mobility array'][idx]

            idx = z['m/z array'] >= min_mz
            z['m/z array'] = z['m/z array'][idx]
            z['intensity array'] = z['intensity array'][idx]
            z['mean inverse reduced ion mobility array'] = z['mean inverse reduced ion mobility array'][idx]

            idx = z['m/z array'] <= max_mz
            z['m/z array'] = z['m/z array'][idx]
            z['intensity array'] = z['intensity array'][idx]
            z['mean inverse reduced ion mobility array'] = z['mean inverse reduced ion mobility array'][idx]

            idx = np.argsort(z['m/z array'])
            z['m/z array'] = z['m/z array'][idx]
            z['intensity array'] = z['intensity array'][idx]
            z['mean inverse reduced ion mobility array'] = z['mean inverse reduced ion mobility array'][idx]

            cnt += 1

            # if len(data_for_analyse) > 5000:
            #     break

            if len(z['m/z array']):
                data_for_analyse.append(z)
            else:
                skipped += 1


    logger.info('Number of MS2 scans: %d', len(data_for_analyse))
    logger.info('Number of skipped MS2 scans: %d', skipped)

    return data_for_analyse



#!/usr/bin/env python

'''
setup.py file for biosaur2
'''
import os
from setuptools import setup, find_packages, Extension

version = open('VERSION').readline().strip()


def make_extensions():
    is_ci = bool(os.getenv("CI", ""))
    include_diagnostics = False
    try:
        import numpy
    except ImportError:
        print("C Extensions require `numpy`")
        raise
    try:
        from Cython.Build import cythonize
        cython_directives = {
            'embedsignature': True,
            "profile": include_diagnostics
        }
        macros = []
        if include_diagnostics:
            macros.append(("CYTHON_TRACE_NOGIL", "1"))
        if is_ci and include_diagnostics:
            cython_directives['linetrace'] = True

        extensions = cythonize([
            Extension(name='biosaur2.cutils', sources=['biosaur2/cutils.pyx'],
                      include_dirs=[numpy.get_include(), ])
        ], compiler_directives=cython_directives)
    except ImportError:
        extensions = [
            Extension(name='biosaur2.cutils', sources=['biosaur2/cutils.c'],
                      include_dirs=[numpy.get_include(), ])
        ]
    return extensions

setup(
    name                 = 'biosaur2',
    version              = version,
    description          = '''A feature detection LC-MS1 spectra.''',
    long_description     = (''.join(open('README.md').readlines())),
    long_description_content_type = 'text/markdown',
    author               = 'Mark Ivanov',
    author_email         = 'markmipt@gmail.com',
    install_requires     = [line.strip() for line in open('requirements.txt')],
    ext_modules          = make_extensions(),
    classifiers          = ['Intended Audience :: Science/Research',
                            'Programming Language :: Python :: 3.9',
                            'Topic :: Education',
                            'Topic :: Scientific/Engineering :: Bio-Informatics',
                            'Topic :: Scientific/Engineering :: Chemistry',
                            'Topic :: Scientific/Engineering :: Physics'],
    license              = 'License :: OSI Approved :: Apache Software License',
    packages             = find_packages(),
    entry_points         = {'console_scripts': ['biosaur2 = biosaur2.search:run',]}
    )



from . import utils
import numpy as np
from scipy.stats import binom
import itertools
from copy import deepcopy
import logging
logger = logging.getLogger(__name__)
from .cutils import get_initial_isotopes, checking_cos_correlation_for_carbon, split_peaks, split_peaks_old, detect_hills, process_hills
from multiprocessing import Queue, Process, cpu_count
from collections import Counter, defaultdict

def split_peaks_python(qout, hills_dict, data_for_analyse_tmp, args, counter_hills_idx, sorted_idx_child_process, sorted_idx_array_child_process, i, checked_id):

    new_index_list = split_peaks(hills_dict, data_for_analyse_tmp, args, counter_hills_idx, sorted_idx_child_process, sorted_idx_array_child_process, i, checked_id)
    qout.put((i, list(new_index_list)))
    qout.put(None)

def split_peaks_multi(hills_dict, data_for_analyse_tmp, args):

    hillValleyFactor = args['hvf']
    min_length_hill = args['minlh']
    # min_length_hill = max(2, min_length_hill)


    hills_dict['orig_idx_array'] = np.array(hills_dict['orig_idx_array'])
    hills_dict['scan_idx_array'] = np.array(hills_dict['scan_idx_array'])
    hills_dict['hills_idx_array'] = np.array(hills_dict['hills_idx_array'])

    counter_hills_idx = Counter(hills_dict['hills_idx_array'])
    counter_hills_idx_2 = dict()
    for k,v in counter_hills_idx.items():
        counter_hills_idx_2[k] = v
    counter_hills_idx = counter_hills_idx_2

    tmp_hill_length = np.array([counter_hills_idx[hill_idx] for hill_idx in hills_dict['hills_idx_array']])
    idx_minl = tmp_hill_length >= min_length_hill
    hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_minl]
    hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_minl]
    hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_minl]

    all_sets = []
    all_sorted_idx = []

    if len(hills_dict['orig_idx_array']):

        idx_sort = np.argsort(hills_dict['hills_idx_array'] + ((hills_dict['scan_idx_array'] + 1) / (hills_dict['scan_idx_array'].max()+2)))
        hills_dict['hills_idx_array'] = hills_dict['hills_idx_array'][idx_sort]
        hills_dict['scan_idx_array'] = hills_dict['scan_idx_array'][idx_sort]
        hills_dict['orig_idx_array'] = hills_dict['orig_idx_array'][idx_sort]

        hills_dict['hills_idx_array_unique'] = sorted(list(set(hills_dict['hills_idx_array'])))

        data_for_analyse_tmp_intensity = [z['intensity array'] for z in data_for_analyse_tmp]

        n_procs = args['nprocs']
        qout = Queue()
        new_idx_res = dict()
        procs = []
        ar2 = []
        len_full = len(hills_dict['hills_idx_array_unique'])
        if len_full <= 1000 * n_procs:
            n_procs = 1
        step = int(len_full / n_procs) + 1
        checked_id = 0
        for i in range(n_procs):
            sorted_idx_child_process = list(hills_dict['hills_idx_array_unique'][i*step:i*step+step])
            idx_unique_set = set(sorted_idx_child_process)
            all_sets.append(idx_unique_set)
            local_idx = np.array([z in idx_unique_set for z in list(hills_dict['hills_idx_array'])])
            sorted_idx_array_child_process = hills_dict['hills_idx_array'][local_idx]
            ar2.append(sorted_idx_array_child_process)
            sorted_idx_child_process = sorted(list(idx_unique_set))

            all_sorted_idx.append(local_idx)
            p = Process(
                target=split_peaks_python,
                args=(qout, hills_dict, data_for_analyse_tmp_intensity, args, counter_hills_idx, sorted_idx_child_process, sorted_idx_array_child_process, i, checked_id))
            checked_id += len(sorted_idx_array_child_process)
                # args=(hills_dict, isotopes_mass_accuracy, isotopes_list, a, min_charge, max_charge, mz_step, paseftol, faims_val, list(sorted_idx_child_process), qout))
            p.start()
            procs.append(p)

        for _ in range(n_procs):
            for ready_child_process in iter(qout.get, None):
                new_idx_res[ready_child_process[0]] = ready_child_process[1]
        for p in procs:
            p.join()

    final_idx_array = []
    last_id = 1
    for i in range(n_procs):
        added_idx_map = {}
        for ii, idx_val in enumerate(new_idx_res[i]):
            if idx_val not in added_idx_map:
                added_idx_map[idx_val] = int(last_id)
                last_id += 1
            final_idx_array.append(added_idx_map[idx_val])

    # hills_dict['scan_idx_array'] = np.concatenate([hills_dict['scan_idx_array'][all_sorted_idx[i]] for i in range(n_procs)])
    # hills_dict['orig_idx_array'] = np.concatenate([hills_dict['orig_idx_array'][all_sorted_idx[i]] for i in range(n_procs)])

    hills_dict['hills_idx_array'] = list(final_idx_array)
    del hills_dict['hills_idx_array_unique']

    return hills_dict

def get_initial_isotopes_python(hills_dict, isotopes_mass_accuracy, isotopes_list, a, min_charge, max_charge, mz_step, paseftol, faims_val, sorted_idx_child_process, qout):

    ready_local = get_initial_isotopes(hills_dict, isotopes_mass_accuracy, isotopes_list, a, min_charge, max_charge, mz_step, paseftol, faims_val, sorted_idx_child_process)
    qout.put(ready_local)
    qout.put(None)

def process_file(args):

    data_for_analyse = utils.process_mzml(args)
    write_header = True

    #Process faims

    faims_set = set([z.get('FAIMS compensation voltage', 0) for z in data_for_analyse])
    if any(z for z in faims_set):
        logger.info('Detected FAIMS values: %s', faims_set)

    data_start_id = 0
    data_cur_id = 0
    RT_dict = dict()

    for faims_val in faims_set:

        if len(faims_set) > 1:
            logger.info('Spectra analysis for CV = %.3f', faims_val)

        data_for_analyse_tmp = []
        for z in data_for_analyse:
            if not faims_val or z['FAIMS compensation voltage'] == faims_val:
                data_for_analyse_tmp.append(z)
                RT_dict[data_cur_id] = float(z['scanList']['scan'][0]['scan start time'])
                data_cur_id += 1


        hill_mass_accuracy = args['htol']
        max_mz_value = 0
        for z in data_for_analyse_tmp:
            max_mz_value = max(max_mz_value, z['m/z array'].max())

        mz_step = hill_mass_accuracy * 1e-6 * max_mz_value

        #Process TOF
        if args['tof']:
            data_for_analyse_tmp = utils.process_tof(data_for_analyse_tmp)

        #Process profile
        if args['profile']:
            data_for_analyse_tmp = utils.process_profile(data_for_analyse_tmp)

        #Process ion mobility

        if all('ignore_ion_mobility' not in z for z in data_for_analyse_tmp):
            utils.centroid_pasef_data(data_for_analyse_tmp, args, mz_step)
        else:
            args['paseftol'] = False

        paseftol = args['paseftol']

        hills_dict = detect_hills(data_for_analyse_tmp, args, mz_step, paseftol)

        hills_dict = split_peaks_multi(hills_dict, data_for_analyse_tmp, args)
        hills_dict = process_hills(hills_dict, data_for_analyse_tmp, mz_step, paseftol, args)
        logger.info('Detected number of hills: %d', len(set(hills_dict['hills_idx_array'])))

        isotopes_mass_accuracy = args['itol']


        isotopes_list = list(range(10))
        averagine_mass = 111.1254
        averagine_C = 4.9384
        a = dict()

        for i in range(100, 20000, 100):
            int_arr = binom.pmf(
                isotopes_list,
                float(i) /
                averagine_mass *
                averagine_C,
                0.0107)
            int_arr_norm = int_arr / int_arr.sum()
            a[i] = int_arr_norm

        min_charge = args['cmin']
        max_charge = args['cmax']

        n_procs = args['nprocs']
        qout = Queue()
        ready = []
        procs = []

        sorted_idx_full = [idx_1 for (idx_1, hill_idx_1), hill_mz_1 in sorted(list(zip(list(enumerate(hills_dict['hills_idx_array_unique'])), hills_dict['hills_mz_median'])), key=lambda x: x[-1])]
        len_full = len(sorted_idx_full)
        step = int(len_full / n_procs)
        for i in range(n_procs):
            sorted_idx_child_process = sorted_idx_full[i*step:i*step+step]

            p = Process(
                target=get_initial_isotopes_python,
                args=(hills_dict, isotopes_mass_accuracy, isotopes_list, a, min_charge, max_charge, mz_step, paseftol, faims_val, list(sorted_idx_child_process), qout))
            p.start()
            procs.append(p)

        for _ in range(n_procs):
            for ready_child_process in iter(qout.get, None):
                ready.extend(ready_child_process)
        for p in procs:
            p.join()


        logger.info('Number of potential isotope clusters: %d', len(ready))

        isotopes_mass_error_map = {}
        for ic in range(1, 10, 1):
            isotopes_mass_error_map[ic] = []

        for i in range(9):
            tmp = []
            for pf in ready:
                isotopes = pf['isotopes']
                if len(isotopes) >= i + 1:
                    tmp.append(isotopes[i]['mass_diff_ppm'])
            isotopes_mass_error_map[i+1] = tmp

        for ic in range(1, 10, 1):
            if ic == 1:

                if len(isotopes_mass_error_map[ic]) >= 100:

                    try:

                        true_md = np.array(isotopes_mass_error_map[ic])

                        mass_left = -min(isotopes_mass_error_map[ic])
                        mass_right = max(isotopes_mass_error_map[ic])


                        mass_shift, mass_sigma, covvalue = utils.calibrate_mass(0.05, mass_left, mass_right, true_md)
                        if abs(mass_shift) >= max(mass_left, mass_right):
                            mass_shift, mass_sigma, covvalue = utils.calibrate_mass(0.25, mass_left, mass_right, true_md)
                        if np.isinf(covvalue):
                            mass_shift, mass_sigma, covvalue = utils.calibrate_mass(0.05, mass_left, mass_right, true_md)

                        isotopes_mass_error_map[ic] = [mass_shift, mass_sigma]

                    except:
                        isotopes_mass_error_map[ic] = [0, 10]

                else:
                    isotopes_mass_error_map[ic] = [0, 10]

            else:
                isotopes_mass_error_map[ic] = deepcopy(isotopes_mass_error_map[ic-1])
                isotopes_mass_error_map[ic][0] = isotopes_mass_error_map[ic][0] - 0.45

        logger.info('Average mass shift between monoisotopic and first 13C isotope: %.3f ppm', isotopes_mass_error_map[1][0])
        logger.info('Average mass std between monoisotopic and first 13C isotope: %.3f ppm', isotopes_mass_error_map[1][1])


        max_l = len(ready)
        cur_l = 0

        while cur_l < max_l:
            pep_feature = ready[cur_l]

            tmp = []

            for cand in pep_feature['isotopes']:
                map_val = isotopes_mass_error_map[cand['isotope_number']]
                if abs(cand['mass_diff_ppm'] - map_val[0]) <= 5 * map_val[1]:
                    tmp.append(cand)
                else:
                    break

            tmp_n_isotopes = len(tmp)

            if tmp_n_isotopes:
                all_theoretical_int, all_exp_intensity = pep_feature['intensity_array_for_cos_corr']
                all_theoretical_int = all_theoretical_int[:tmp_n_isotopes+1]
                all_exp_intensity = all_exp_intensity[:tmp_n_isotopes+1]
                cos_corr, number_of_passed_isotopes = checking_cos_correlation_for_carbon(all_theoretical_int, all_exp_intensity, 0.6)
                if cos_corr:

                    ready[cur_l]['cos_cor_isotopes'] = cos_corr
                    ready[cur_l]['isotopes'] = tmp
                    ready[cur_l]['nIsotopes'] = tmp_n_isotopes + 1
                    ready[cur_l]['intensity_array_for_cos_corr'] = [all_theoretical_int, all_exp_intensity]
                    # ready[cur_l]['sumI'] = np.log10(sum(all_exp_intensity))
                    # ready[cur_l]['mass_diff_ppm_abs'] = abs(ready[cur_l]['isotopes'][0]['mass_diff_ppm'])

                    

                else:
                    del ready[cur_l]
                    max_l -= 1
                    cur_l -= 1


            else:
                del ready[cur_l]
                max_l -= 1
                cur_l -= 1

            cur_l += 1

        logger.info('Number of potential isotope clusters after smart mass accuracy for isotopes: %d', len(ready))

        max_l = len(ready)
        cur_l = 0

        func_for_sort = lambda x: -x['nIsotopes']-x['cos_cor_isotopes']
        # func_for_sort = lambda x: -x['nIsotopes']*1e6+x['mass_diff_ppm_abs']

        ready_final = []
        ready_set = set()
        # ready_set = defaultdict(int)
        ready = sorted(ready, key=func_for_sort)
        cur_isotopes = ready[0]['nIsotopes']


        cnt_mark = 0

        while cur_l < max_l:
            cnt_mark += 1
            pep_feature = ready[cur_l]
            n_iso = pep_feature['nIsotopes']
            if n_iso < cur_isotopes:
                ready = sorted(ready, key=func_for_sort)
                cur_isotopes = n_iso
                cur_l = 0

            if pep_feature['monoisotope hill idx'] not in ready_set:
                if not any(cand['isotope_hill_idx'] in ready_set for cand in pep_feature['isotopes']):
                # if not any(ready_set[cand['isotope_hill_idx']]>1 for cand in pep_feature['isotopes']):
                    ready_final.append(pep_feature)
                    ready_set.add(pep_feature['monoisotope hill idx'])
                    # ready_set[pep_feature['monoisotope hill idx']] += 1
                    for cand in pep_feature['isotopes']:
                        ready_set.add(cand['isotope_hill_idx'])
                        # ready_set[cand['isotope_hill_idx']] += 1
                    del ready[cur_l]
                    max_l -= 1
                    cur_l -= 1

                else:
                    tmp = []

                    for cand in pep_feature['isotopes']:
                        if cand['isotope_hill_idx'] not in ready_set:
                        # if ready_set[cand['isotope_hill_idx']] <= 1:
                            tmp.append(cand)
                        else:
                            break

                    tmp_n_isotopes = len(tmp)

                    if tmp_n_isotopes:
                        all_theoretical_int, all_exp_intensity = pep_feature['intensity_array_for_cos_corr']
                        all_theoretical_int = all_theoretical_int[:tmp_n_isotopes+1]
                        all_exp_intensity = all_exp_intensity[:tmp_n_isotopes+1]
                        cos_corr, number_of_passed_isotopes = checking_cos_correlation_for_carbon(all_theoretical_int, all_exp_intensity, 0.6)
                        if cos_corr:

                            ready[cur_l]['cos_cor_isotopes'] = cos_corr
                            ready[cur_l]['isotopes'] = tmp
                            ready[cur_l]['nIsotopes'] = tmp_n_isotopes + 1
                            ready[cur_l]['intensity_array_for_cos_corr'] = [all_theoretical_int, all_exp_intensity]
                            # ready[cur_l]['sumI'] = np.log10(sum(all_exp_intensity))
                            # ready[cur_l]['mass_diff_ppm_abs'] = abs(ready[cur_l]['isotopes'][0]['mass_diff_ppm'])

                        else:
                            del ready[cur_l]
                            max_l -= 1
                            cur_l -= 1


                    else:
                        del ready[cur_l]
                        max_l -= 1
                        cur_l -= 1
            else:
                del ready[cur_l]
                max_l -= 1
                cur_l -= 1

            cur_l += 1

        logger.info('Number of detected isotope clusters: %d', len(ready_final))


        negative_mode = args['nm']
        peptide_features = utils.calc_peptide_features(hills_dict, ready_final, negative_mode, faims_val, RT_dict, data_start_id)

        utils.write_output(peptide_features, args, write_header)

        write_header = False

        data_start_id += len(data_for_analyse_tmp)
